from dataclasses import dataclass
from typing import List

from ..api.noun_phrase import NounPhrase
from .types import ConceptType


@dataclass
class QuestionDescriptor:
    noun_phrases: List[NounPhrase]
    type: ConceptType



from .book_authentication_descriptor import BookAuthenticationDescriptor
from .book_custom_authentication_descriptor import \
    BookCustomAuthenticationDescriptor
from .book_oauth_authentication_descriptor import (
    BookOAuthAuthenticationDescriptor, OAuthFlow, OAuthProvider,
    OAuthTokenFunction)
from .credential_descriptor import CredentialDescriptor
from .oauth_argument_descriptor import OauthArgumentDescriptor



from __future__ import annotations

from dataclasses import dataclass

from .concept_descriptor import ConceptDescriptor


@dataclass
class BookConfigDescriptor:
    python_property_name: str
    concept: ConceptDescriptor



import inspect
from typing import List, Optional, Type

from ..api.noun_phrase import NounPhrase
from .authentication import (BookAuthenticationDescriptor,
                             BookOAuthAuthenticationDescriptor,
                             OAuthTokenFunction)
from .book_config_descriptor import BookConfigDescriptor
from .book_procedure_descriptor import BookProcedureDescriptor


class BookDescriptor:
    cls: Type
    _id: str
    _name: str
    _author: Optional[str]
    _short_description: Optional[str]
    _long_description: Optional[str]
    _icon: Optional[bytes]
    _configuration: List[BookConfigDescriptor]

    def __init__(  # pylint: disable=too-many-arguments
        self,
        cls: Type,
        identifier: str,
        name: Optional[str] = None,
        noun_phrase: Optional[NounPhrase] = None,
        author: Optional[str] = None,
        short_description: Optional[str] = None,
        long_description: Optional[str] = None,
        icon: Optional[bytes] = None,
        icon_path: Optional[str] = None,
        configuration: Optional[List[BookConfigDescriptor]] = None,
    ):
        self._id = identifier
        self._name = name
        self.cls = cls
        self._author = author
        self._noun_phrase = noun_phrase
        self._short_description = short_description
        self._long_description = long_description
        self._icon = icon
        self._icon_path = icon_path
        self._configuration = configuration

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> Optional[str]:
        return self._name

    @property
    def noun_phrase(self) -> Optional[NounPhrase]:
        return self._noun_phrase

    @property
    def author(self) -> Optional[str]:
        return self._author

    @property
    def short_description(self) -> Optional[str]:
        return self._short_description

    @property
    def long_description(self) -> Optional[str]:
        return self._long_description

    @property
    def configuration(self) -> List[BookConfigDescriptor]:
        return self._configuration

    @property
    def icon(self) -> Optional[bytes]:
        return self._icon

    @property
    def icon_path(self) -> Optional[str]:
        return self._icon_path

    @property
    def authentications(self) -> List[BookAuthenticationDescriptor]:
        authentications = []
        oauth_token_function = None

        for _, member in inspect.getmembers(self.cls, predicate=inspect.isfunction):
            authentication = getattr(member, "__connect__", None)
            if authentication:
                authentications.append(authentication)

            is_oauth_token_function = getattr(member, "__oauthtoken__", None)
            if is_oauth_token_function:
                oauth_token_function = OAuthTokenFunction(name=member.__name__, access_token_is_sensitive=member.__access_token_is_sensitive__)

        oauths: List[BookOAuthAuthenticationDescriptor] = getattr(self.cls, "__oauth__", [])

        for oauth in oauths:
            oauth.oauth_token_function = oauth_token_function

        authentications.extend(oauths)

        return authentications

    @property
    def procedures(self) -> List[BookProcedureDescriptor]:
        has_authentications = len(self.authentications) > 0
        procedures = []

        for _, member in inspect.getmembers(self.cls, predicate=inspect.isfunction):
            procedure = getattr(member, "__procedure__", None)
            if procedure:
                procedure.connection_required = procedure.override_connection_required if procedure.override_connection_required is not None else has_authentications
                procedures.append(procedure)

        return procedures

    @property
    def blueprints(self) -> List[str]:
        return getattr(self.cls, "__blueprints__", [])

    @property
    def connection_required(self) -> bool:
        procedures = self.procedures
        return all(p.connection_required for p in procedures)

    @property
    def discover_function_name(self) -> Optional[str]:
        for _, member in inspect.getmembers(self.cls, predicate=inspect.isfunction):
            if getattr(member, "__discover__", None):
                return member.__name__

        return None

    @property
    def invoke_function_name(self) -> Optional[str]:
        for _, member in inspect.getmembers(self.cls, predicate=inspect.isfunction):
            if getattr(member, "__invoke__", None):
                return member.__name__

        return None

    @property
    def discover_capable(self) -> bool:
        return self.discover_function_name is not None and self.invoke_function_name is not None



from .blueprint_decorator import blueprint, blueprint_procedure
from .book_decorator import book
from .concept_decorator import concept
from .config_decorator import config
from .connect_decorator import connect
from .discover_decorator import discover
from .invoke_decorator import invoke
from .oauth_decorator import OAuthFlow, OAuthProvider, oauth
from .oauthtoken_decorator import oauthtoken
from .procedure_decorator import procedure



from docstring_parser import ParseError
from docstring_parser.google import GoogleParser

from .docstring import Docstring
from .error import DocstringParseError
from .sections import DEFAULT_SECTIONS


class DocstringParser:  # pylint disable=too-few-public-methods
    @classmethod
    def parse(cls, text: str) -> Docstring:
        try:
            parser = GoogleParser(DEFAULT_SECTIONS)
            return Docstring(docstring=parser.parse(text))
        except ParseError as ex:
            raise DocstringParseError(original_exception=ex) from ex



import inspect

from kognitos.bdk.typing import Sensitive


def oauthtoken(fn):
    if not inspect.isfunction(fn):
        raise TypeError("The oauthtoken decorator can only be applied to functions.")

    sig = inspect.signature(fn)
    parameters = sig.parameters
    if "access_token" not in parameters:
        raise ValueError("The method decorated with @oauthtoken must have a parameter named 'access_token'")
    access_token_param = parameters["access_token"]

    if hasattr(access_token_param.annotation, "__origin__") and access_token_param.annotation.__origin__ == Sensitive:
        fn.__access_token_is_sensitive__ = True
    else:
        fn.__access_token_is_sensitive__ = False
    fn.__oauthtoken__ = True

    return fn



class DocstringParseError(RuntimeError):

    def __init__(self, message="Cannot parse docstring", original_exception=None):
        """Initialize the exception with an optional error message."""
        self.message = message
        self.original_exception = original_exception

        if self.original_exception:
            super().__init__(f"{self.message}: {str(self.original_exception)}")
        else:
            super().__init__(self.message)



import importlib
import inspect
import sys
import uuid
from dataclasses import MISSING, fields, is_dataclass
from datetime import date, datetime, time
from decimal import Decimal
from enum import Enum, EnumMeta
from functools import lru_cache
from importlib.metadata import PackageNotFoundError
from types import NoneType, UnionType
from typing import (IO, Any, ForwardRef, List, Optional, Tuple, Union,
                    get_origin)

from ...api.noun_phrase import NounPhrase
from ...docstring import DocstringParser
from ...typing import Sensitive
from ..types import ConceptTableType
from ..types.any import ConceptAnyType
from ..types.base import ConceptType
from ..types.dict import ConceptDictionaryType, ConceptDictionaryTypeField
from ..types.enum import ConceptEnumType, ConceptEnumTypeMember
from ..types.list import ConceptListType
from ..types.opaque import ConceptOpaqueType
from ..types.optional import ConceptOptionalType
from ..types.scalar import ConceptScalarType
from ..types.self import ConceptSelfType
from ..types.sensitive import ConceptSensitiveType
from ..types.union import ConceptUnionType


@lru_cache
def is_attrs_installed():
    try:
        importlib.metadata.distribution("attrs")
        return True
    except PackageNotFoundError:
        return False


@lru_cache
def is_pyarrow_installed():
    try:
        importlib.metadata.distribution("pyarrow")
        return True
    except PackageNotFoundError:
        return False


@lru_cache
def is_arro3_installed():
    try:
        importlib.metadata.distribution("arro3.core")
        return True
    except PackageNotFoundError:
        return False


@lru_cache
def is_nanoarrow_installed():
    try:
        importlib.metadata.distribution("nanoarrow")
        return True
    except PackageNotFoundError:
        return False


if is_attrs_installed():
    from .attrs_utils import from_attrs, is_attrs


if is_pyarrow_installed():
    from pyarrow import Table as PyArrowTable
else:
    PyArrowTable = None

if is_arro3_installed():
    from arro3.core import \
        Table as Arro3Table  # pylint: disable=no-name-in-module
else:
    Arro3Table = None

if is_nanoarrow_installed():
    from nanoarrow import ArrayStream as NanoArrowArrayStream
else:
    NanoArrowArrayStream = None


def should_translate_to_optional(annotation: type, unset: Optional[Any] = None):
    if not get_origin(annotation) in (Union, UnionType):
        return False
    unset_types = [NoneType] if unset is None else [NoneType, unset.__class__]
    unset_types_present = list(map(lambda t: t in unset_types, annotation.__args__)).count(True)
    return unset_types_present >= 1


def get_truthy_types(annotation: type, unset: Optional[Any]) -> List[Any]:
    unset_types = [NoneType] if unset is None else [NoneType, unset.__class__]
    return [t for t in annotation.__args__ if t not in unset_types]


class ConceptTypeFactory:

    @classmethod
    def from_type(cls, annotation: type, backward: Optional[List[str]] = None, unset: Optional[Any] = None) -> ConceptType:
        from ...api.questions import Question  # pylint: disable=cyclic-import

        if getattr(getattr(annotation, "__origin__", None), "__name__", None) == "Question":
            raise ValueError(
                "Questions cannot be used as a concept type. If you are type-hinting a procedure to indicate that it can return a question, use a union type at the top level of the return type-hint. If the procedure can return a question, but has no results, use 'None | Question' as the return type-hint."
            )

        if annotation is Question:
            raise ValueError(
                "Question typehints must be parameterized: Question[Literal['noun phrase'], type]"
            )

        if backward is None:
            backward = []

        if isinstance(annotation, EnumMeta):
            is_a = set(getattr(annotation, "__is_a__", []))

            docstring = DocstringParser.parse(annotation.__doc__ or "")

            keys = sorted(list(annotation.__members__.keys()))
            resolved_members = [docstring.enum_member_by_name(key) for key in keys]

            members = [ConceptEnumTypeMember(member.name, member.description, member.noun_phrase) for member in resolved_members]

            return ConceptEnumType(is_a=is_a, members=members, description=docstring.short_description or docstring.long_description, concrete=annotation)

        if annotation == str:
            return ConceptScalarType.TEXT
        if annotation in (int, float, Decimal):
            return ConceptScalarType.NUMBER
        if annotation == bool:
            return ConceptScalarType.BOOLEAN
        if annotation == datetime:
            return ConceptScalarType.DATETIME
        if annotation == date:
            return ConceptScalarType.DATE
        if annotation == time:
            return ConceptScalarType.TIME
        if annotation == uuid.UUID:
            return ConceptScalarType.UUID
        if should_translate_to_optional(annotation, unset):
            truthy_types = get_truthy_types(annotation, unset)
            if len(truthy_types) == 1:
                return ConceptOptionalType(ConceptTypeFactory.from_type(truthy_types[0], backward, unset)).simplify()

            inner_types = [ConceptTypeFactory.from_type(t, backward, unset) for t in truthy_types]
            return ConceptOptionalType(ConceptUnionType(inner_types)).simplify()
        if get_origin(annotation) == Sensitive:
            return ConceptSensitiveType(ConceptTypeFactory.from_type(annotation.__args__[0], backward, unset)).simplify()
        if get_origin(annotation) == list:
            return ConceptListType(ConceptTypeFactory.from_type(annotation.__args__[0], backward, unset)).simplify()
        if get_origin(annotation) in (Union, UnionType):
            inner_types = [ConceptTypeFactory.from_type(arg, backward, unset) for arg in annotation.__args__]
            return ConceptUnionType(inner_types).simplify()
        if get_origin(annotation) == IO or (inspect.isclass(annotation) and issubclass(annotation, IO)):
            return ConceptScalarType.FILE
        if annotation == NounPhrase:
            return ConceptScalarType.CONCEPTUAL
        if annotation == Any:
            return ConceptAnyType()
        if inspect.isclass(annotation) and issubclass(annotation, Enum):
            return ConceptTypeFactory.from_type(type(next(iter(annotation)).value), unset=unset)
        if isinstance(annotation, ForwardRef):
            if annotation.__forward_arg__ in backward:
                return ConceptSelfType()

            globalns, localns = collect_namespaces()
            resolved = annotation._evaluate(globalns, localns, recursive_guard=set())

            return ConceptTypeFactory.from_type(resolved, backward + [annotation.__forward_arg__], unset=unset).simplify()

        if is_dataclass(annotation):
            return from_dataclass(annotation, unset)

        if is_attrs_installed():
            if is_attrs(annotation):  # pyright: ignore [reportPossiblyUnboundVariable]
                return from_attrs(annotation, cls, unset)  # pyright: ignore [reportPossiblyUnboundVariable]

        if is_pyarrow_installed():
            if PyArrowTable and annotation == PyArrowTable:
                return ConceptTableType(None, [], PyArrowTable)

        if is_arro3_installed():
            if Arro3Table and annotation == Arro3Table:
                return ConceptTableType(None, [], Arro3Table)

        if is_nanoarrow_installed():
            if NanoArrowArrayStream and annotation == NanoArrowArrayStream:
                return ConceptTableType(None, [], NanoArrowArrayStream)

        if hasattr(annotation, "__is_a__"):
            return from_serializable(annotation)

        if get_origin(annotation) == dict:
            return ConceptDictionaryType(set(), None, None, [])

        return ConceptOpaqueType({NounPhrase.from_head("thing")}, None, annotation)


def compute_default_value(field: Any) -> Tuple[bool, Any]:
    if field.default is not MISSING:
        return True, field.default

    if field.default_factory is not MISSING:
        return True, field.default_factory()

    return False, None


def from_dataclass(annotation, unset: Optional[Any] = None):
    docstring = DocstringParser.parse(annotation.__doc__)
    unset = getattr(annotation, "__unset__", unset)
    dict_fields = []

    for field in fields(annotation):
        description: Optional[str] = next((attribute.description for attribute in docstring.attributes if attribute.name == field.name), None)

        has_default_value, default_value = compute_default_value(field)

        if field.init is False:
            concept_type = ConceptTypeFactory.from_type(Optional[field.type], unset=unset)  # type: ignore
        else:
            concept_type = ConceptTypeFactory.from_type(field.type, unset=unset)  # type: ignore

        dict_fields.append(ConceptDictionaryTypeField(field.name, description, concept_type, default_value=default_value, has_default_value=has_default_value, init=field.init))

    return ConceptDictionaryType(
        set(getattr(annotation, "__is_a__", [])),
        annotation,
        (
            (docstring.short_description or "") + (docstring.long_description or "")
            if docstring.short_description and docstring.long_description
            else docstring.short_description or docstring.long_description or ""
        ),
        dict_fields,
        unset=unset,
    )


def from_serializable(annotation):
    docstring = DocstringParser.parse(annotation.__doc__)

    return ConceptOpaqueType(
        set(getattr(annotation, "__is_a__")),
        (
            (docstring.short_description or "") + (docstring.long_description or "")
            if docstring.short_description and docstring.long_description
            else docstring.short_description or docstring.long_description or ""
        ),
        annotation,
    )


def collect_namespaces():
    globalns = {}
    localns = {}

    frame = sys._getframe()
    while frame:
        globalns.update(frame.f_globals)
        localns.update(frame.f_locals)
        frame = frame.f_back

    return globalns, localns



from .book import BookFactory
from .concept import ConceptFactory
from .config import BookConfigFactory
from .custom_authentication import BookCustomAuthenticationFactory
from .parameter_concept import ParameterConceptFactory
from .procedure import BookProcedureFactory
from .types import ConceptTypeFactory



import base64
from abc import ABC, abstractmethod
from typing import Any, Dict, Generic, List, Optional, Type, TypeVar, Union

from ..reflection.types.base import ConceptType
from .noun_phrase import NounPhrase

T = TypeVar("T")
NounPhrasesStringLiteral = TypeVar("NounPhrasesStringLiteral", bound=str)


class Question(Generic[NounPhrasesStringLiteral, T]):
    id: str
    noun_phrases: List[NounPhrase]
    concept_type: ConceptType
    choices: List[T]
    text: Optional[str]

    @staticmethod
    def generate_question_id(noun_phrases: NounPhrasesStringLiteral, identifier: str = "") -> str:
        concat = (noun_phrases + identifier).encode("utf-8")
        return base64.b64encode(concat).decode("utf-8")  # NOTE: using base64 instead of hash because the rust's PyO3 yields inconsistent results with it

    def __init__(self, noun_phrases: NounPhrasesStringLiteral, concept_type: Type[T], choices: Optional[List[T]] = None, text: Optional[str] = None, identifier: str = ""):
        """
        Initialize a Question object.

        Args:
            noun_phrases: The name of the concept whose value you want to ask for.
            concept_type: The type that you expect the answer to be.
            choices: The choices for the user to select (if any).
            text: The text you want to present to the user (if any).
            identifier: If for some reason you need to ask for the same
                noun phrases twice, but you expect different answers, use
                this as a 'nonce' to differentiate between them.
        """
        from ..reflection.factory.types import ConceptTypeFactory

        if not isinstance(noun_phrases, str):
            raise TypeError("noun_phrases must be a string literal")
        self.id = self.generate_question_id(noun_phrases, identifier)
        self.noun_phrases = [NounPhrase.from_str(noun_phrase) for noun_phrase in noun_phrases.split("'s ")]
        self.concept_type = ConceptTypeFactory.from_type(concept_type)
        self.choices = choices if choices else []
        self.text = text

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Question) and self.id == other.id

    def __str__(self) -> str:
        ret = "Question("
        nps_text = "'s ".join(str(np) for np in self.noun_phrases)
        if self.text:
            ret += f"{self.text}, "

        ret += f"noun_phrases='{nps_text}'"

        if self.choices:
            ret += f", choices={self.choices}"

        return ret + ")"

    def __repr__(self) -> str:
        return str(self)


class AnswerStorage(ABC):
    @abstractmethod
    def get(self, key: str) -> Optional[Any]:
        raise NotImplementedError()

    @abstractmethod
    def set(self, key: str, answer_value: Any):
        raise NotImplementedError()

    @abstractmethod
    def clear(self):
        raise NotImplementedError()

    @abstractmethod
    def unset(self, key: str, fail_if_not_set: bool):
        raise NotImplementedError()


class InMemoryAnswerStorage(AnswerStorage):
    def __init__(self):
        self._answer_map: Dict[str, Any] = {}

    def get(self, key: str) -> Optional[Any]:
        return self._answer_map.get(key)

    def set(self, key: str, answer_value: Any):
        self._answer_map[key] = answer_value

    def unset(self, key: str, fail_if_not_set: bool):
        if key in self._answer_map:
            del self._answer_map[key]
        elif fail_if_not_set:
            raise ValueError(f"Key {key} is not set")

    def clear(self):
        self._answer_map.clear()


# Use the InMemoryAnswerStorage by default
answer_storage = InMemoryAnswerStorage()


def get_from_context(key: str) -> Optional[Any]:
    return answer_storage.get(key)


def set_answer(key: str, answer_value: Any):
    answer_storage.set(key, answer_value)


def unset_answer(key: str, fail_if_not_set: bool = False):
    answer_storage.unset(key, fail_if_not_set)


def clear_answers():
    answer_storage.clear()


def ask(
    concept_name: NounPhrasesStringLiteral, concept_type: Type[T], text: Optional[str] = None, choices: Optional[List[T]] = None, identifier: str = ""
) -> Union[T, Question[NounPhrasesStringLiteral, T]]:
    """
    Ask a question and get an answer (if available) or a Question object to be
    returned to the user.

    Args:
        concept_name: The name of the concept whose value you want to ask for.
        concept_type: The type that you expect the answer to be.
        choices: The choices for the user to select (if any).
        text: The text you want to present to the user (if any).
        identifier: If for some reason you need to ask for the same
            noun phrases twice, but you expect different answers, use
            this as a 'nonce' to differentiate between them.

    Returns:
        Answer containing the value from the answer map, typed as T (the
        `concept_type`) or a Question (typed as `Question[Literal[concept_name], T]`)
        if the question is not answered yet.
    """

    key = Question.generate_question_id(concept_name, identifier)
    return get_from_context(key) or Question(concept_name, concept_type, choices, text, identifier)



from dataclasses import dataclass
from enum import Enum
from typing import List, Optional

from .book_authentication_descriptor import BookAuthenticationDescriptor
from .oauth_argument_descriptor import OauthArgumentDescriptor


class OAuthFlow(Enum):
    AUTHORIZATION_CODE = 0
    CLIENT_CREDENTIALS = 1


class OAuthProvider(Enum):
    MICROSOFT = 0
    GOOGLE = 1
    GENERIC = 2


@dataclass
class OAuthTokenFunction:
    name: str
    access_token_is_sensitive: bool


@dataclass
class BookOAuthAuthenticationDescriptor(BookAuthenticationDescriptor):
    provider: OAuthProvider
    flows: List[OAuthFlow]
    authorize_endpoint: str
    token_endpoint: str
    scopes: Optional[List[str]]
    arguments: Optional[List[OauthArgumentDescriptor]] = None
    oauth_token_function: Optional[OAuthTokenFunction] = None



import inspect
from functools import wraps
from inspect import Signature

from ..docstring import DocstringParser
from ..reflection import BookProcedureSignature
from ..reflection.factory import BookProcedureFactory


def procedure(name: str, **kwargs):
    override_connection_required = kwargs.get("connection_required", None)

    def decorator(fn):
        if not inspect.isfunction(fn):
            raise TypeError("The procedure decorator can only be applied to functions.")

        # parse procedure signature
        english_signature = BookProcedureSignature.from_english(name)

        # parse python signature
        python_signature = extract_python_signature(fn)

        # parse documentation
        if not fn.__doc__:
            raise ValueError("missing docstring")

        docstring = DocstringParser.parse(fn.__doc__)

        # construct book_procedure
        book_procedure = BookProcedureFactory.create(fn.__name__, english_signature, python_signature, docstring, override_connection_required)

        if not hasattr(fn, "__procedure__"):
            fn.__procedure__ = book_procedure

        if not hasattr(fn, "__signature__"):
            fn.__signature__ = python_signature

        if not hasattr(fn, "__text_signature__"):
            fn.__text_signature__ = english_signature

        return wraps(fn)(fn)

    def extract_python_signature(fn) -> Signature:
        return inspect.signature(fn, eval_str=True)

    return decorator



from .any import ConceptAnyType
from .base import ConceptType
from .credential import CredentialScalarType, CredentialType
from .dict import ConceptDictionaryType, ConceptDictionaryTypeField
from .enum import ConceptEnumType, ConceptEnumTypeMember
from .list import ConceptListType
from .opaque import ConceptOpaqueType
from .optional import ConceptOptionalType
from .scalar import ConceptScalarType
from .self import ConceptSelfType
from .sensitive import ConceptSensitiveType
from .table import ConceptTableType, ConceptTableTypeColumn
from .union import ConceptUnionType



from typing import List, Optional, Set, Type

from kognitos.bdk.api.noun_phrase import NounPhrase

from .base import ConceptType


class ConceptEnumTypeMember:
    name: str
    description: Optional[str]
    noun_phrase: Optional[NounPhrase]

    def __eq__(self, other) -> bool:
        if isinstance(other, ConceptEnumTypeMember):
            return self.name == other.name and self.description == other.description and self.noun_phrase == other.noun_phrase

        return False

    def __init__(self, name: str, description: Optional[str] = None, noun_phrase: Optional[NounPhrase] = None):
        self.name = name
        self.description = description
        self.noun_phrase = noun_phrase


class ConceptEnumType(ConceptType):
    is_a: Set[NounPhrase]
    members: List[ConceptEnumTypeMember]
    description: Optional[str]
    concrete: Optional[Type]

    def __init__(self, is_a: Set[NounPhrase], members: List[ConceptEnumTypeMember], description: Optional[str] = None, concrete: Optional[Type] = None):
        self.is_a = is_a
        self.members = members
        self.description = description
        self.concrete = concrete

    def simplify(self) -> ConceptType:
        return self

    def children(self) -> List[ConceptType]:
        return []

    def __repr__(self) -> str:
        return f"Enum[{', '.join(self.keys)}]"

    def __str__(self) -> str:
        return self.__repr__()

    def __eq__(self, other) -> bool:
        if isinstance(other, ConceptEnumType):
            return self.keys == other.keys and self.is_a == other.is_a
        return False

    def __hash__(self) -> int:
        return hash("".join(self.keys))

    def replace(self, old: ConceptType, new: ConceptType) -> ConceptType:
        if self == old:
            return new
        return self

    @property
    def keys(self) -> List[str]:
        return [m.name for m in self.members]



import inspect
from functools import wraps


def invoke(fn):
    if not inspect.isfunction(fn):
        raise TypeError("The invoke decorator can only be applied to functions.")

    fn.__invoke__ = True
    return wraps(fn)(fn)
